@@ -4,19 +4,24 @@ use strict;
 use warnings;
 
 use version;
-our $VERSION = version->new('0.01');
+our $VERSION = version->new('0.02');
 
 use Carp;
 use Try::Tiny;
 
 use FindBin qw($Script);
+use File::Copy;
 
 use base qw/ App::ClusterSSH::Base /;
 use App::ClusterSSH::Cluster;
 
 my $clusters;
 my %old_clusters;
-my @app_specific   = (qw/ command title comms method ssh rsh telnet ccon /);
+my @app_specific = (qw/ command title comms method parent /);
+
+# list of config items to not write out when writing the default config
+my @ignore_default_config = (qw/ user /);
+
 my %default_config = (
     terminal                   => "xterm",
     terminal_args              => "",
@@ -28,15 +33,19 @@ my %default_config = (
     terminal_size              => "80x24",
 
     use_hotkeys             => "yes",
-    key_quit                => "Control-q",
+    key_quit                => "Alt-q",
     key_addhost             => "Control-Shift-plus",
     key_clientname          => "Alt-n",
     key_history             => "Alt-h",
+    key_localname           => "Alt-l",
     key_retilehosts         => "Alt-r",
+    key_macros_enable       => "Alt-p",
     key_paste               => "Control-v",
+    key_username            => "Alt-u",
     mouse_paste             => "Button-2",
     auto_quit               => "yes",
     auto_close              => 5,
+    use_natural_sort        => 0,
     window_tiling           => "yes",
     window_tiling_direction => "right",
     console_position        => "",
@@ -54,11 +63,22 @@ my %default_config = (
     terminal_decoration_height => 10,
     terminal_decoration_width  => 8,
 
-    rsh_args    => "",
-    telnet_args => "",
-    ssh_args    => "",
+    console      => 'console',
+    console_args => '',
+    rsh          => 'rsh',
+    rsh_args     => "",
+    telnet       => 'telnet',
+    telnet_args  => "",
+    ssh          => 'ssh',
+    ssh_args     => "",
+    sftp         => 'sftp',
+    sftp_args    => "",
 
-    extra_cluster_file => "",
+    extra_tag_file           => '',
+    extra_cluster_file       => '',
+    external_cluster_command => '',
+    external_command_mode    => '0600',
+    external_command_pipe    => '',
 
     unmap_on_redraw => "no",    # Debian #329440
 
@@ -67,15 +87,34 @@ my %default_config = (
     history_height => 10,
 
     command             => q{},
+    hide_menu           => 0,
     max_host_menu_items => 30,
 
+    macros_enabled   => 'yes',
+    macro_servername => '%s',
+    macro_hostname   => '%h',
+    macro_username   => '%u',
+    macro_newline    => '%n',
+    macro_version    => '%v',
+
     max_addhost_menu_cluster_items => 6,
     menu_send_autotearoff          => 0,
     menu_host_autotearoff          => 0,
 
+    unique_servers    => 0,
     use_all_a_records => 0,
 
-    send_menu_xml_file => $ENV{HOME} . '/.csshrc_send_menu',
+    send_menu_xml_file => $ENV{HOME} . '/.clusterssh/send_menu',
+
+    auto_wm_decoration_offsets => "no",    # Debian #842965
+
+    # don't set username here as takes precendence over ssh config
+    user => '',
+    rows => -1,
+    cols => -1,
+
+    fillscreen => "no",
+
 );
 
 sub new {
@@ -84,21 +123,20 @@ sub new {
     my $self = $class->SUPER::new(%default_config);
 
     ( my $comms = $Script ) =~ s/^c//;
-    $self->{comms} = $comms;
+
+    $comms = 'telnet'  if ( $comms eq 'tel' );
+    $comms = 'console' if ( $comms eq 'con' );
+    $comms = 'ssh'     if ( $comms eq 'lusterssh' );
+    $comms = 'sftp'    if ( $comms eq 'sftp' );
 
     # list of allowed comms methods
-    if ( 'ssh rsh telnet console' !~ m/\B$comms\B/ ) {
+    if ( 'ssh rsh telnet sftp console' !~ m/\b$comms\b/ ) {
         $self->{comms} = 'ssh';
     }
-
-    if ( $self->{comms}
-        && ( !$self->{ $self->{comms} } || !-e $self->{ $self->{comms} } ) )
-    {
-        $self->{ $self->{comms} } = $self->find_binary( $self->{comms} );
+    else {
+        $self->{comms} = $comms;
     }
 
-    $self->{terminal} = $self->find_binary( $self->{terminal} );
-
     $self->{title} = uc($Script);
 
     $clusters = App::ClusterSSH::Cluster->new();
@@ -131,13 +169,41 @@ sub validate_args {
             App::ClusterSSH::Exception::Config->throw(
                 unknown_config => \@unknown_config,
                 error          => $self->loc(
-                    'Unknown configuration parameters: [_1]',
+                    'Unknown configuration parameters: [_1]' . $/,
                     join( ',', @unknown_config )
                 )
             )
         );
     }
 
+    if ( !$self->{comms} ) {
+        croak(
+            App::ClusterSSH::Exception::Config->throw(
+                error => $self->loc( 'Invalid variable: comms' . $/ ),
+            ),
+        );
+    }
+
+    if ( !$self->{ $self->{comms} } ) {
+        croak(
+            App::ClusterSSH::Exception::Config->throw(
+                error => $self->loc(
+                    'Invalid variable: [_1]' . $/,
+                    $self->{comms}
+                ),
+            ),
+        );
+    }
+
+    # check the terminal has been found correctly
+    # looking for the terminal should not be fatal
+    if ( !-e $self->{terminal} ) {
+        eval { $self->{terminal} = $self->find_binary( $self->{terminal} ); };
+        if ($@) {
+            warn $@->message;
+        }
+    }
+
     return $self;
 }
 
@@ -146,48 +212,54 @@ sub parse_config_file {
 
     $self->debug( 2, 'Loading in config file: ', $config_file );
 
-    if ( !-e $config_file || !-r $config_file ) {
-        croak(
-            App::ClusterSSH::Exception::Config->throw(
-                error => $self->loc(
-                    'File [_1] does not exist or cannot be read', $config_file
-                ),
-            ),
-        );
-    }
+    #    if ( !-e $config_file || !-r $config_file ) {
+    #        croak(
+    #            App::ClusterSSH::Exception::Config->throw(
+    #                error => $self->loc(
+    #                    'File [_1] does not exist or cannot be read' . $/,
+    #                    $config_file
+    #                ),
+    #            ),
+    #        );
+    #    }
+    #
+    #    open( CFG, $config_file ) or die("Couldnt open $config_file: $!");
+    #    my $l;
+    #    my %read_config;
+    #    while ( defined( $l = <CFG> ) ) {
+    #        next
+    #            if ( $l =~ /^\s*$/ || $l =~ /^#/ )
+    #            ;    # ignore blank lines & commented lines
+    #        $l =~ s/#.*//;     # remove comments from remaining lines
+    #        $l =~ s/\s*$//;    # remove trailing whitespace
+    #
+    #        # look for continuation lines
+    #        chomp $l;
+    #        if ( $l =~ s/\\\s*$// ) {
+    #            $l .= <CFG>;
+    #            redo unless eof(CFG);
+    #        }
+    #
+    #        next unless $l =~ m/\s*(\S+)\s*=\s*(.*)\s*/;
+    #        my ( $key, $value ) = ( $1, $2 );
+    #        if ( defined $key && defined $value ) {
+    #            $read_config{$key} = $value;
+    #            $self->debug( 3, "$key=$value" );
+    #        }
+    #    }
+    #    close(CFG);
 
-    open( CFG, $config_file ) or die("Couldnt open $config_file: $!");
-    my $l;
     my %read_config;
-    while ( defined( $l = <CFG> ) ) {
-        next
-            if ( $l =~ /^\s*$/ || $l =~ /^#/ )
-            ;    # ignore blank lines & commented lines
-        $l =~ s/#.*//;     # remove comments from remaining lines
-        $l =~ s/\s*$//;    # remove trailing whitespace
+    %read_config
+        = $self->load_file( type => 'config', filename => $config_file );
 
-        # look for continuation lines
-        chomp $l;
-        if ( $l =~ s/\\\s*$// ) {
-            $l .= <CFG>;
-            redo unless eof(CFG);
-        }
-
-        next unless $l =~ m/\s*(\S+)\s*=\s*(.*)\s*/;
-        my ( $key, $value ) = ( $1, $2 );
-        if ( defined $key && defined $value ) {
-            $read_config{$key} = $value;
-            $self->debug( 3, "$key=$value" );
-        }
-    }
-    close(CFG);
-
-    # grab any c'lusters from the config before validating it
+    # grab any clusters from the config before validating it
     if ( $read_config{clusters} ) {
         $self->debug( 3, "Picked up clusters defined in $config_file" );
         foreach my $cluster ( sort split / /, $read_config{clusters} ) {
             if ( $read_config{$cluster} ) {
-                $clusters->register_tag( $cluster, $read_config{$cluster} );
+                $clusters->register_tag( $cluster,
+                    split( / /, $read_config{$cluster} ) );
                 $old_clusters{$cluster} = $read_config{$cluster};
                 delete( $read_config{$cluster} );
             }
@@ -205,16 +277,6 @@ sub parse_config_file {
 sub load_configs {
     my ( $self, @configs ) = @_;
 
-    if ( -e $ENV{HOME} . '/.csshrc' ) {
-        warn(
-            $self->loc(
-                'NOTICE: [_1] is no longer used - please see documentation and remove',
-                $ENV{HOME} . '/.csshrc'
-            ),
-            $/
-        );
-    }
-
     for my $config (
         '/etc/csshrc',
         $ENV{HOME} . '/.csshrc',
@@ -248,6 +310,30 @@ sub load_configs {
 sub write_user_config_file {
     my ($self) = @_;
 
+    # attempt to move the old config file to one side
+    if ( -f "$ENV{HOME}/.csshrc" ) {
+        eval { move( "$ENV{HOME}/.csshrc", "$ENV{HOME}/.csshrc.DISABLED" ) };
+
+        if ($@) {
+            croak(
+                App::ClusterSSH::Exception::Config->throw(
+                    error => $self->loc(
+                        'Unable to move [_1] to [_2]: [_3]' . $/,
+                        '$HOME/.csshrc', '$HOME/.csshrc.DISABLED', $@
+                    ),
+                )
+            );
+        }
+        else {
+            warn(
+                $self->loc(
+                    'Moved [_1] to [_2]' . $/, '$HOME/.csshrc',
+                    '$HOME/.csshrc.DISABLED'
+                ),
+            );
+        }
+    }
+
     return if ( -f "$ENV{HOME}/.clusterssh/config" );
 
     if ( !-d "$ENV{HOME}/.clusterssh" ) {
@@ -255,7 +341,7 @@ sub write_user_config_file {
             croak(
                 App::ClusterSSH::Exception::Config->throw(
                     error => $self->loc(
-                        'Unable to create directory [_1]: [_2]',
+                        'Unable to create directory [_1]: [_2]' . $/,
                         '$HOME/.clusterssh', $!
                     ),
                 ),
@@ -264,37 +350,80 @@ sub write_user_config_file {
         }
     }
 
+    # Debian #673507 - migrate clusters prior to writing ~/.clusterssh/config
+    # in order to update the extra_cluster_file property
+    if (%old_clusters) {
+        if ( open( my $fh, ">", "$ENV{HOME}/.clusterssh/clusters" ) ) {
+            print $fh '# '
+                . $self->loc('Tag definitions moved from old .csshrc file'),
+                $/;
+            foreach ( sort( keys(%old_clusters) ) ) {
+                print $fh $_, ' ', join( ' ', $old_clusters{$_} ), $/;
+            }
+            close($fh);
+        }
+        else {
+            croak(
+                App::ClusterSSH::Exception::Config->throw(
+                    error => $self->loc(
+                        'Unable to write [_1]: [_2]' . $/,
+                        '$HOME/.clusterssh/clusters',
+                        $!
+                    ),
+                ),
+            );
+        }
+    }
+
     if ( open( CONFIG, ">", "$ENV{HOME}/.clusterssh/config" ) ) {
         foreach ( sort( keys(%$self) ) ) {
-            print CONFIG "$_=$self->{$_}\n";
+            my $comment = '';
+            if ( grep /$_/, @ignore_default_config ) {
+                $comment = '#';
+            }
+            print CONFIG ${comment}, $_, '=', $self->{$_}, $/;
         }
         close(CONFIG);
+        warn(
+            $self->loc(
+                'Created new configuration file within [_1]' . $/,
+                '$HOME/.clusterssh/'
+            )
+        );
     }
     else {
         croak(
             App::ClusterSSH::Exception::Config->throw(
                 error => $self->loc(
-                    'Unable to write default [_1]: [_2]',
-                    '$HOME/.clusterssh/config',
-                    $!
+                    'Unable to write default [_1]: [_2]' . $/,
+                    '$HOME/.clusterssh/config', $!
                 ),
             ),
         );
     }
 
-    return $self if ( !%old_clusters );
-
-    if ( open( my $fh, ">", "$ENV{HOME}/.clusterssh/clusters" ) ) {
-        print $fh '# '
-            . $self->loc('Tag definitions moved from old .csshrc file'), $/;
-        foreach ( sort( keys(%old_clusters) ) ) {
-            print $fh $_, ' ', join( ' ', $old_clusters{$_} ), $/;
-        }
-        close($fh);
-    }
     return $self;
 }
 
+# search given directories for the given file
+sub search_dirs {
+    my ( $self, $file, @directories ) = @_;
+
+    my $path;
+
+    foreach my $dir (@directories) {
+        $self->debug( 3, "Looking for $file in $dir" );
+
+        if ( -f $dir . '/' . $file && -x $dir . '/' . $file ) {
+            $path = $dir . '/' . $file;
+            $self->debug( 2, "Found at $path" );
+            last;
+        }
+    }
+
+    return $path;
+}
+
 # could use File::Which for some of this but we also search a few other places
 # just in case $PATH isnt set up right
 sub find_binary {
@@ -303,7 +432,7 @@ sub find_binary {
     if ( !$binary ) {
         croak(
             App::ClusterSSH::Exception::Config->throw(
-                error => $self->loc('argument not provided'),
+                error => $self->loc('argument not provided') . $/,
             ),
         );
     }
@@ -313,39 +442,39 @@ sub find_binary {
     # if not found, strip the path and look again
     if ( $binary =~ m!^/! ) {
         if ( -f $binary ) {
-            $self->debug( 2, "$binary already fully qualified" );
+            $self->debug( 2, "Already have full path to in $binary" );
             return $binary;
         }
         else {
-            $self->debug( 2, "$binary not found - re-searching" );
+            $self->debug( 2, "Full path for $binary incorrect; searching" );
             $binary =~ s!^.*/!!;
         }
     }
 
     my $path;
     if ( !-x $binary || substr( $binary, 0, 1 ) ne '/' ) {
+        $path = $self->search_dirs( $binary, split( /:/, $ENV{PATH} ) );
 
-        foreach (
-            split( /:/, $ENV{PATH} ), qw!
-            /bin
-            /sbin
-            /usr/sbin
-            /usr/bin
-            /usr/local/bin
-            /usr/local/sbin
-            /opt/local/bin
-            /opt/local/sbin
-            !
-            )
-        {
-            $self->debug( 3, "Looking in $_" );
-
-            if ( -f $_ . '/' . $binary && -x $_ . '/' . $binary ) {
-                $path = $_ . '/' . $binary;
-                $self->debug( 2, "Found at $path" );
-                last;
-            }
+        # if it is on $PATH then no need to qualitfy the path to it
+        # keep it as it is
+        if ($path) {
+            return $binary;
         }
+        else {
+            $path = $self->search_dirs(
+                $binary, qw!
+                    /bin
+                    /sbin
+                    /usr/sbin
+                    /usr/bin
+                    /usr/local/bin
+                    /usr/local/sbin
+                    /opt/local/bin
+                    /opt/local/sbin
+                    !
+            );
+        }
+
     }
     else {
         $self->debug( 2, "Already configured OK" );
@@ -355,7 +484,8 @@ sub find_binary {
         croak(
             App::ClusterSSH::Exception::Config->throw(
                 error => $self->loc(
-                    '"[_1]" binary not found - please amend $PATH or the cssh config file',
+                    '"[_1]" binary not found - please amend $PATH or the cssh config file'
+                        . $/,
                     $binary
                 ),
             ),
@@ -370,13 +500,17 @@ sub dump {
     my ( $self, $no_exit, ) = @_;
 
     $self->debug( 3, 'Dumping config to STDOUT' );
-    print( '# Configuration dump produced by "cssh -u"', $/ );
+    print( '# Configuration dump produced by "cssh -d"', $/ );
 
     foreach my $key ( sort keys %$self ) {
+        my $comment = '';
         if ( grep /$key/, @app_specific ) {
             next;
         }
-        print $key, '=', $self->{$key}, $/;
+        if ( grep /$key/, @ignore_default_config ) {
+            $comment = '#';
+        }
+        print $comment, $key, '=', $self->{$key}, $/;
     }
 
     $self->exit if ( !$no_exit );
@@ -396,7 +530,7 @@ sub dump {
 
 =head1 NAME
 
-ClusterSSH::Config
+ClusterSSH::Config - Object representing application configuration
 
 =head1 SYNOPSIS
 
@@ -420,6 +554,10 @@ Read in configuration from given filename
 
 Validate and apply all configuration loaded at this point
 
+=item $path = $config->search_dirs('<name>', @seaarch_directories);
+
+Search the given directories for the name given.  Return undef if not found.
+
 =item $path = $config->find_binary('<name>');
 
 Locate the binary <name> and return the full path.  Doesn't just search 
@@ -447,7 +585,7 @@ Duncan Ferguson, C<< <duncan_j_ferguson at yahoo.co.uk> >>
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 1999-2010 Duncan Ferguson.
+Copyright 1999-2016 Duncan Ferguson.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
